import json
import os
from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from scan_models.tests.constances import standard_output


class TestCommand(TestCase):
    maxDiff = None

    def call_command(self, *args, **kwargs):
        out = StringIO()
        call_command(
            "scan_models",
            *args,
            stdout=out,
            stderr=StringIO(),
            **kwargs,
        )
        return out.getvalue()

    def test_all_models(self):
        self.call_command()

        with open(os.path.abspath("test_output/output.json"), "r") as file:
            self.assertEqual(standard_output, json.load(file))

    def test_location_prefix(self):
        self.call_command(prefix="test_output")

        # Double test output because extra prefix
        with open(os.path.abspath("test_output/test_output/output.json"), "r") as file:
            self.assertEqual(standard_output, json.load(file))



from typing import Type

from scan_models.settings import get_setting
from scan_models.validators.base_validator import BaseValidator
from pydoc import locate


class Factory:
    @staticmethod
    def get_validator() -> Type[BaseValidator]:
        validator = locate(get_setting("validator"))

        if not issubclass(validator, BaseValidator):
            raise TypeError("Validator is not of type BaseValidator")

        return validator



from collections import OrderedDict

from django.core.validators import MaxValueValidator, MinValueValidator, RegexValidator
from django.db.models import fields

from scan_models.factory import Factory


class ValidatorParser:
    def __init__(self, field: fields.Field):
        self.field = field
        self.validator = OrderedDict()
        self.validator_class = Factory.get_validator()()

    def parse(self):
        self._calculate_required()
        self._calculate_max_length()
        self._calculate_choices()
        self._calculate_max_min_value()
        self._calculate_email()
        self._calculate_regex()

        return self.validator

    def _calculate_required(self):
        required = self.field.default == fields.NOT_PROVIDED and not self.field.blank and not self.field.null

        if required:
            self.validator_class.set_required(self.validator)

    def _calculate_email(self):
        if isinstance(self.field, fields.EmailField):
            self.validator_class.set_is_email(self.validator)

    def _calculate_max_length(self):
        if self.field.max_length:
            self.validator_class.set_max_length(self.validator, self.field.max_length)

    def _calculate_choices(self):
        if self.field.choices:
            self.validator_class.set_choices(self.validator, [choice[0] for choice in self.field.choices])

    def _calculate_regex(self):
        regex_validator = self._find_validator(RegexValidator)

        if regex_validator:
            self.validator_class.set_regex(self.validator, regex_validator.regex.pattern)

    def _calculate_max_min_value(self):
        if not isinstance(self.field, fields.IntegerField):
            return

        max_validator = self._find_validator(MaxValueValidator)
        min_validator = self._find_validator(MinValueValidator)

        internal_type = self.field.get_internal_type()
        min_value, max_value = fields.connection.ops.integer_field_range(internal_type)

        if max_validator and max_validator.limit_value != max_value:
            self.validator_class.set_max_value(self.validator, max_validator.limit_value)

        if min_validator and min_validator.limit_value != min_value:
            self.validator_class.set_min_value(self.validator, min_validator.limit_value)

    def _find_validator(self, validator_class):
        validator = [validator for validator in self.field.validators if isinstance(validator, validator_class)]

        return validator[0] if len(validator) else None



from enum import Enum

from scan_models.settings import get_setting


class Verbosity(Enum):
    ONE = 1
    TWO = 2


def is_verbosity(verbosity: Verbosity):
    return verbosity.value <= get_setting("verbosity")



from django.db.models import fields

from scan_models.parser.attributes import AttributesParser
from scan_models.parser.validator import ValidatorParser
from scan_models.validators.base_validator import BaseValidator
from scan_models.verbosity import is_verbosity, Verbosity


class FieldParser:
    field: fields.Field
    validator_class: BaseValidator

    def __init__(self, field: fields.Field):
        self.field = field

        super().__init__()

    def parse(self):
        # If auto field there is no frontend validation
        if isinstance(self.field, fields.AutoField):
            return None

        parsed = {}

        validator = ValidatorParser(self.field).parse()
        if validator or is_verbosity(Verbosity.ONE):
            parsed["validator"] = validator

        attributes = AttributesParser(self.field).parse()
        if attributes or is_verbosity(Verbosity.ONE):
            parsed["attributes"] = attributes

        return parsed



import uuid
from collections import OrderedDict
from unittest.mock import MagicMock

from django.conf import settings
from django.db.models import fields, Choices
from django.test import TestCase

from scan_models.parser.attributes import AttributesParser
from scan_models.settings import DEFAULT_SETTINGS

od = OrderedDict()


class TestAttributesParser(TestCase):
    def test_nothing(self):
        self.assertEqual(od, AttributesParser(fields.CharField()).parse())

    def test_function_calls(self):
        parser = AttributesParser(fields.CharField())
        parser._calculate_options = MagicMock()
        parser._calculate_element = MagicMock()
        parser._calculate_type = MagicMock()
        parser._calculate_default = MagicMock()

        parser.parse()

        self.assertEqual(parser._calculate_options.call_count, 1)
        self.assertEqual(parser._calculate_element.call_count, 1)
        self.assertEqual(parser._calculate_type.call_count, 1)
        self.assertEqual(parser._calculate_default.call_count, 1)

    def test_type(self):
        # Number type
        parser = AttributesParser(fields.IntegerField())
        parser._calculate_type()
        self.assertEqual("number", parser.attributes["type"])

        # Email type
        parser = AttributesParser(fields.EmailField())
        parser._calculate_type()
        self.assertEqual("email", parser.attributes["type"])

    def test_element(self):
        # Need verbosity 2 for elements in attributes
        settings.SCAN_MODELS["verbosity"] = 2

        # Checkbox
        parser = AttributesParser(fields.BooleanField())
        parser._calculate_element()
        self.assertEqual("checkbox", parser.attributes["element"])

        # TextArea
        parser = AttributesParser(fields.TextField())
        parser._calculate_element()
        self.assertEqual("textarea", parser.attributes["element"])

        # Date
        parser = AttributesParser(fields.DateField())
        parser._calculate_element()
        self.assertEqual("date", parser.attributes["element"])

        # Select
        parser = AttributesParser(fields.CharField(choices=(("1", "1"), ("2", "2"))))
        parser._calculate_element()
        self.assertEqual("select", parser.attributes["element"])

        # Reset verbosity
        settings.SCAN_MODELS["verbosity"] = DEFAULT_SETTINGS["verbosity"]

    def test_options(self):
        class TestChoices(Choices):
            YES = "wat"
            NO = "test"

        parser = AttributesParser(fields.CharField(choices=TestChoices.choices))
        parser._calculate_options()
        self.assertEqual(["Yes", "No"], parser.attributes["options"])

    def test_default(self):
        parser = AttributesParser(fields.CharField())
        parser._calculate_default()
        self.assertFalse("default" in parser.attributes)

        parser = AttributesParser(fields.CharField(default=""))
        parser._calculate_default()
        self.assertFalse("default" in parser.attributes)

        parser = AttributesParser(fields.CharField(default="test"))
        parser._calculate_default()
        self.assertEqual("test", parser.attributes["default"])

        parser = AttributesParser(fields.CharField(default=uuid.uuid4()))
        parser._calculate_default()
        self.assertFalse("default" in parser.attributes)



from collections import OrderedDict

from django.db.models import fields

from scan_models.verbosity import is_verbosity, Verbosity


class AttributesParser:
    def __init__(self, field: fields.Field):
        self.field = field
        self.attributes = OrderedDict()

    def parse(self):
        self._calculate_element()
        self._calculate_type()
        self._calculate_options()
        self._calculate_default()

        return self.attributes

    def _calculate_element(self):
        # Only calculate element if verbosity is high enough. This is only the case when using the general-fields package
        if not is_verbosity(Verbosity.TWO):
            return

        element = ""

        if self.field.choices:
            element = "select"
        elif isinstance(self.field, fields.TextField):
            element = "textarea"
        elif isinstance(self.field, fields.BooleanField):
            element = "checkbox"
        elif isinstance(self.field, fields.DateField):
            element = "date"

        if element:
            self.attributes["element"] = element

    def _calculate_type(self):
        type_attr = ""
        if isinstance(self.field, fields.IntegerField):
            type_attr = "number"
        elif isinstance(self.field, fields.EmailField):
            type_attr = "email"

        if type_attr:
            self.attributes["type"] = type_attr

    def _calculate_options(self):
        if self.field.choices:
            self.attributes["options"] = [choice[1] for choice in self.field.choices]

    def _calculate_default(self):
        """
        Only adds primitive types as default

        TODO: Add more options as default
        """

        if (
            self.field.default
            and self.field.default != fields.NOT_PROVIDED
            and isinstance(self.field.default, (list, int, str, bool, float))
        ):
            self.attributes["default"] = str(self.field.default)



import json
import os

from scan_models.main import scan_model

standard_output = {
    "name": {"validator": {"max": 30}, "attributes": {}},
    "email": {"validator": {"required": True, "max": 254, "email": True}, "attributes": {"type": "email"}},
    "max_amount": {"validator": {"required": True, "max_value": 4}, "attributes": {"type": "number"}},
    "min_amount": {"validator": {"required": True, "min_value": 1}, "attributes": {"type": "number"}},
    "choices": {
        "validator": {"required": True, "max": 4, "oneOf": ["yes", "no"]},
        "attributes": {"options": ["Yes", "No"], "element": "select"},
    },
    "a_or_b": {"validator": {"max": 200, "regex": "(a|b)"}, "attributes": {}},
}


def create_test(model):
    path = "./output.json"
    scan_model(model, path)

    file = open(os.path.abspath(path), "r")
    data = json.load(file)

    return data



import json
import os
from pathlib import Path

from django.apps import apps

from scan_models.parser.main import FieldParser
from scan_models.settings import get_setting


def scan_model(model_name: str, output: str, location_prefix: str = ""):
    """
    Scan the model and output it to a file

    Args:
        model_name (str): The name of the model. Preferably with the app. Example: companies.Company
        output (str): Output file. Advised to use .json on the end
    """

    fields = apps.get_model(model_name)._meta.fields

    validator = {}

    for field in fields:
        validator_field = FieldParser(field).parse()

        if validator_field:
            validator[_format_name(field.name)] = validator_field

    # We need to make sure the directory where we put the file exists
    location_prefix = Path(location_prefix or "")
    output_path = location_prefix.joinpath(Path(output))
    if not output_path.parent.exists():
        output_path.parent.mkdir(parents=True)

    with open(output_path.resolve(), "w") as outfile:
        json.dump(validator, outfile, indent=2)


def _format_name(value: str):
    """
    Format the name of the field

    Args:
        value: The string to format
    """

    if get_setting("camelize"):
        components = value.split("_")
        return components[0] + "".join(x.title() for x in components[1:])
    else:
        return value



from django.core.management import BaseCommand

from scan_models.main import scan_model
from scan_models.settings import get_setting


class Command(BaseCommand):
    help = "Creates the hour registration for the last week"

    def add_arguments(self, parser):
        parser.add_argument(
            "-m",
            "--model",
            help="If you only want to specify one model",
        )
        parser.add_argument(
            "-p",
            "--prefix",
            help="Prefix for the location of the mapping",
        )

    def handle(self, *args, **kwargs):
        data = get_setting("mapping")
        option_model = kwargs.get("model")
        location_prefix = kwargs.get("prefix")

        if option_model:
            scan_model(option_model, data[option_model], location_prefix)
        else:
            for model, output in data.items():
                scan_model(model, output, location_prefix)



from copy import copy

from django.conf import settings
from django.test import TestCase

from scan_models.settings import DEFAULT_SETTINGS
from scan_models.tests.constances import standard_output, create_test


class TestHappyFlow(TestCase):
    maxDiff = None

    def setUp(self) -> None:
        settings.SCAN_MODELS = copy(DEFAULT_SETTINGS)

    def test_happy_flow(self):
        data = create_test("tests.TestModel")
        self.assertEqual(data, standard_output)

    def test_camelize(self):
        settings.SCAN_MODELS["camelize"] = True
        output = copy(standard_output)

        output["maxAmount"] = output.pop("max_amount")
        output["minAmount"] = output.pop("min_amount")
        output["aOrB"] = output.pop("a_or_b")

        data = create_test("tests.TestModel")
        self.assertEqual(data, output)



from django.conf import settings
from django.test import TestCase

from scan_models.settings import DEFAULT_SETTINGS
from scan_models.tests.constances import create_test


class TestVerbosity(TestCase):
    def test_lowest_verbosity(self):
        self.assertEqual(create_test("tests.TestVerbosity"), {})

    def test_first_verbosity(self):
        settings.SCAN_MODELS["verbosity"] = 1
        self.assertEqual(create_test("tests.TestVerbosity"), {"text": {"attributes": {}, "validator": {}}})
        settings.SCAN_MODELS["verbosity"] = DEFAULT_SETTINGS["verbosity"]

    def test_second_verbosity(self):
        settings.SCAN_MODELS["verbosity"] = 2
        self.assertEqual(
            create_test("tests.TestVerbosity"), {"text": {"attributes": {"element": "textarea"}, "validator": {}}}
        )
        settings.SCAN_MODELS["verbosity"] = DEFAULT_SETTINGS["verbosity"]



from collections import OrderedDict
from unittest.mock import MagicMock

from django.core.validators import MinValueValidator, MaxValueValidator, RegexValidator, URLValidator
from django.db import models
from django.test import TestCase

from scan_models.parser.validator import ValidatorParser

od = OrderedDict()


class TestValidatorParser(TestCase):
    def test_skip_auto(self):
        parser = ValidatorParser(models.AutoField())
        self.assertFalse(bool(parser.parse()))

    def test_required(self):
        parser = ValidatorParser(models.CharField())
        set_required = MagicMock()
        parser.validator_class.set_required = set_required

        # Test basic CharField
        parser._calculate_required()
        self.assertTrue(set_required.called)
        set_required.reset_mock()

        # Check with null set to true
        parser.field = models.CharField(null=True)
        parser._calculate_required()
        self.assertFalse(set_required.called)
        set_required.reset_mock()

        # Check with blank set to true
        parser.field = models.CharField(blank=True)
        parser._calculate_required()
        self.assertFalse(set_required.called)
        set_required.reset_mock()

        # Check with a default set
        parser.field = models.CharField(default="")
        parser._calculate_required()
        self.assertFalse(set_required.called)
        set_required.reset_mock()

    def test_max_length(self):
        parser = ValidatorParser(models.CharField(max_length=44))
        set_max_length = MagicMock()
        parser.validator_class.set_max_length = set_max_length

        parser._calculate_max_length()
        set_max_length.assert_called_with(od, 44)
        set_max_length.reset_mock()

        parser.field = models.IntegerField()
        parser._calculate_max_length()
        self.assertFalse(set_max_length.called)
        set_max_length.reset_mock()

    def test_choices(self):
        class ChoiceClass(models.Choices):
            FIRST = "first"
            SECOND = "second"

        parser = ValidatorParser(models.CharField(choices=ChoiceClass.choices))
        set_choices = MagicMock()
        parser.validator_class.set_choices = set_choices

        # Test with choices
        parser._calculate_choices()
        set_choices.assert_called_with(od, ["first", "second"])
        set_choices.reset_mock()

        parser.field = models.CharField()
        parser._calculate_choices()
        self.assertFalse(set_choices.called)
        set_choices.reset_mock()

    def test_email(self):
        parser = ValidatorParser(models.CharField())
        set_is_email = MagicMock()
        parser.validator_class.set_is_email = set_is_email

        parser._calculate_email()
        self.assertFalse(set_is_email.called)
        set_is_email.reset_mock()

        parser.field = models.EmailField()
        parser._calculate_email()
        self.assertTrue(set_is_email.called)
        set_is_email.reset_mock()

    def test_regex(self):
        parser = ValidatorParser(models.CharField())
        set_regex = MagicMock()
        parser.validator_class.set_regex = set_regex

        parser._calculate_regex()
        self.assertFalse(set_regex.called)
        set_regex.reset_mock()

        parser.field = models.CharField(validators=[RegexValidator(regex=r"(a|b)")])

        parser._calculate_regex()
        self.assertTrue(set_regex.called)
        set_regex.reset_mock()

        parser.field = models.CharField(validators=[URLValidator()])

        parser._calculate_regex()
        self.assertTrue(set_regex.called)
        set_regex.reset_mock()

    def test_max_min_value(self):
        parser = ValidatorParser(models.CharField())
        set_max_value = MagicMock()
        parser.validator_class.set_max_value = set_max_value

        set_min_value = MagicMock()
        parser.validator_class.set_min_value = set_min_value

        parser._calculate_max_min_value()
        self.assertFalse(set_max_value.called)
        set_max_value.reset_mock()
        self.assertFalse(set_min_value.called)
        set_min_value.reset_mock()

        parser.field = models.IntegerField()
        parser._calculate_max_min_value()
        self.assertFalse(set_max_value.called)
        set_max_value.reset_mock()
        self.assertFalse(set_min_value.called)
        set_min_value.reset_mock()

        parser.field = models.IntegerField(validators=[MinValueValidator(3), MaxValueValidator(5)])
        parser._calculate_max_min_value()
        set_max_value.assert_called_with(od, 5)
        set_max_value.reset_mock()
        set_min_value.assert_called_with(od, 3)
        set_min_value.reset_mock()
